#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include "NLKernel_classify.h"

TEjemplo tabla_ejemplos[MAXEJEMPLOS],tabla_modelo[MAXEJEMPLOS];
int num_ejemplos,num_ejemplos_modelo,num_dimensiones;
double gamm;
double Y[MAXEJEMPLOS];

main(int argc, char *argv[])
{
  int i,j,l,errores;
  char car1,car2;

  /* Comprobamos correccion de la linea de comandos. */
  if(argc!=3) {
    printf("USO: NLKernel_classify fich_modelo fich_pruebas\n");
    exit(0);
  }
 
  /* Leemos el fichero del modelo en memoria: */
  leer_modelo(argv[1]);
  
  printf("%d ejemplos del modelo, %d dimensiones\n",num_ejemplos_modelo,num_dimensiones);
  /* Imprimimos la maquina leida: */
  /*printf("%f\n",gamm);
  for(i=0;i<num_ejemplos_modelo;i++)
    printf("Y[%d]=%f; ",i,Y[i]);
  printf("\n");
  for(i=0;i<num_ejemplos_modelo;i++){
    printf("Clase:%d -> ",tabla_modelo[i].clase);
    for(j=0;j<num_dimensiones;j++){
      printf("%f ",tabla_modelo[i].vector[j]);
    }
    printf("\n");
  }*/

  /* Leemos el fichero de ejemplos en memoria: */
  leer_ejemplos(argv[2]);
  
  printf("%d ejemplos a testear, %d dimensiones\n",num_ejemplos,num_dimensiones);
  /* Imprimimos los ejemplos a probar. */
  /*for(i=0;i<num_ejemplos;i++){
    printf("Clase:%d -> ",tabla_ejemplos[i].clase);
    for(j=0;j<num_dimensiones;j++){
      printf("%f ",tabla_ejemplos[i].vector[j]);
    }
    printf("\n");
  }*/

  /* Comenzamos a clasificar ejemplos del fichero de entrada con el modelo
   * cargado en memoria (con el metodo de Kernel pseudo-inversa) .
   */
  errores = 0;
  for(i=0;i<num_ejemplos;i++)
    if(tabla_ejemplos[i].clase !=  KPIClassify(tabla_ejemplos[i].vector))
      errores++;
  printf("%d/%d ejemplos mal clasificados -> %.2f%% aciertos.\n",
         errores,num_ejemplos,100-100*(double)errores/num_ejemplos);
}

void leer_ejemplos(char *fichero)
{
  int i,clase,atributo;
  double valor;
  FILE *f;

  f=fopen(fichero,"r");
  if(f==NULL) {
    perror("Error en apertura de fichero de ejemplos.");
    exit(1);
  }
  fscanf(f,"%d",&clase);
  num_ejemplos=0;
  num_dimensiones=0;
  while(1) {
    tabla_ejemplos[num_ejemplos].clase = clase;
    for(i=0;i<MAXDIM;i++)
      tabla_ejemplos[num_ejemplos].vector[i] = 0.0;   
    while(fscanf(f,"%d:%lg ",&atributo,&valor)==2){
      tabla_ejemplos[num_ejemplos].vector[atributo-1] = valor;
      if(atributo>num_dimensiones)
        num_dimensiones = atributo;
    }
    num_ejemplos++;
    if(feof(f)){
      break;
    }
    clase=atributo;
  }
  fclose(f);
}

void leer_modelo(char *fichero)
{
  int i,clase,atributo;
  double valor;
  FILE *f;

  f=fopen(fichero,"r");
  if(f==NULL) {
    perror("Error en apertura de fichero de modelo.");
    exit(1);
  }

  /* Leemos el valor de gamma: */
  fscanf(f,"%lg",&gamm);
  /* Leemos el vector Y:*/
  fscanf(f,"%d",&num_ejemplos_modelo);
  for(i=0;i<num_ejemplos_modelo;i++) {
    fscanf(f,"%lg ",&Y[i]);
  }
  /* Y ahora comenzamos a leer los ejemplos: */
  num_ejemplos_modelo = 0;
  num_dimensiones = 0;
  fscanf(f,"%d",&clase);
  while(1) {
    tabla_modelo[num_ejemplos_modelo].clase = clase;
    for(i=0;i<MAXDIM;i++)
      tabla_modelo[num_ejemplos_modelo].vector[i] = 0.0;   
    while(fscanf(f,"%d:%lg ",&atributo,&valor)==2){
      tabla_modelo[num_ejemplos_modelo].vector[atributo-1] = valor;
      if(atributo>num_dimensiones)
        num_dimensiones=atributo;
    }
    num_ejemplos_modelo++;
    if(feof(f)){
      break;
    }
    clase=atributo;
  }
  fclose(f);
}


double kernel(double *v1,double *v2)
{
  int i;
  double acum=0;
  
  for(i=0;i<num_dimensiones;i++) {
    acum += (v1[i]-v2[i])*(v1[i]-v2[i]);
  }
  return exp(-gamm*acum);
}

double dot(double *v1, double *v2)
{
  int i;
  double acum=0;
  
  for(i=0;i<num_ejemplos_modelo;i++)
    acum += v1[i]*v2[i];
  return acum;
}

int KPIClassify(double *v)
{
  double r[MAXEJEMPLOS];
  int i;
  
  for(i=0;i<num_ejemplos_modelo;i++) {
    r[i] = kernel(v,tabla_modelo[i].vector);
  }
  if(dot(r,Y)>0) {
    return 1;
  } else {
    return -1;
  }
}

