#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include "Kernel_Mahalanobis_classify.h"

TEjemplo tabla_ejemplos[MAXEJEMPLOS],tabla_modelo[MAXEJEMPLOS];

int num_ejemplos,num_ejemplos_modelo,num_ejemplos_sep[2],num_reduc_sep[2],
    num_dimensiones;
double gamm;
double K[2][MAXEJEMPLOSSEP][MAXEJEMPLOSSEP];
double W[3],UNOK[2][MAXEJEMPLOSSEP];

int main(int argc, char *argv[])
{
  int i,l,errores;

  /* Comprobamos correccion de la linea de comandos. */
  if(argc!=3) {
    printf("USO: NLKernel_classify fich_modelo fich_pruebas\n");
    exit(0);
  }
 
  /* Leemos el fichero del modelo en memoria: */
  leer_modelo(argv[1]);
  
  printf("%d ejemplos del modelo, %d dimensiones\n",num_ejemplos_modelo,num_dimensiones);
  for(l=0;l<2;l++) {
    printf("K%d (%dx%d)\n",l,num_ejemplos_sep[l],num_ejemplos_sep[l]);
  }
  /* Imprimimos la maquina leida: */
  /*printf("%f\n",gamm);
  for(i=0;i<num_ejemplos_modelo;i++)
    printf("Y[%d]=%f; ",i,Y[i]);
  printf("\n");
  for(i=0;i<num_ejemplos_modelo;i++){
    printf("Clase:%d -> ",tabla_modelo[i].clase);
    for(j=0;j<num_dimensiones;j++){
      printf("%f ",tabla_modelo[i].vector[j]);
    }
    printf("\n");
  }*/

  /* Leemos el fichero de ejemplos en memoria: */
  leer_ejemplos(argv[2]);
  
  printf("%d ejemplos a testear, %d dimensiones\n",num_ejemplos,num_dimensiones);
  /* Imprimimos los ejemplos a probar. */
  /*for(i=0;i<num_ejemplos;i++){
    printf("Clase:%d -> ",tabla_ejemplos[i].clase);
    for(j=0;j<num_dimensiones;j++){
      printf("%f ",tabla_ejemplos[i].vector[j]);
    }
    printf("\n");
  }*/

  /* Comenzamos a clasificar ejemplos del fichero de entrada con el modelo
   * cargado en memoria (con el metodo de distancia a la media Kernel.
   */
  errores = 0;
  for(i=0;i<num_ejemplos;i++)
    if(tabla_ejemplos[i].clase !=  KMahClassify(tabla_ejemplos[i].vector))
      errores++;
  printf("%d/%d ejemplos mal clasificados -> %.2f%% aciertos.\n",
         errores,num_ejemplos,100-100*(double)errores/num_ejemplos);
 
  return 0;
}

void leer_ejemplos(char *fichero)
{
  int i,clase,atributo;
  double valor;
  FILE *f;

  f=fopen(fichero,"r");
  if(f==NULL) {
    perror("Error en apertura de fichero de ejemplos.");
    exit(1);
  }
  fscanf(f,"%d",&clase);
  num_ejemplos=0;
  num_dimensiones=0;
  while(1) {
    tabla_ejemplos[num_ejemplos].clase = clase;
    for(i=0;i<MAXDIM;i++)
      tabla_ejemplos[num_ejemplos].vector[i] = 0.0;   
    while(fscanf(f,"%d:%lg ",&atributo,&valor)==2){
      tabla_ejemplos[num_ejemplos].vector[atributo-1] = valor;
      if(atributo>num_dimensiones)
        num_dimensiones = atributo;
    }
    num_ejemplos++;
    if(feof(f)){
      break;
    }
    clase=atributo;
  }
  fclose(f);
}

void leer_modelo(char *fichero)
{
  int i,j,l,clase,atributo;
  double valor;
  FILE *f;

  f=fopen(fichero,"r");
  if(f==NULL) {
    perror("Error en apertura de fichero de modelo.");
    exit(1);
  }

  /* Leemos el valor de gamma: */
  fscanf(f,"%lg",&gamm);
  /* Leemos el separador lineal: */
  fscanf(f,"%lg",&(W[0]));
  fscanf(f,"%lg",&(W[1]));
  fscanf(f,"%lg",&(W[2]));
  for(l=0;l<2;l++) {
    /* Leemos el nmero de ejemplos:*/
    fscanf(f,"%d",&(num_ejemplos_sep[l]));
    /* Leemos el numero de filas de la matriz: */
    fscanf(f,"%d",&(num_reduc_sep[l]));
    /* Leemos el vector mu-x */
    for(i=0;i<num_ejemplos_sep[l];i++) 
      fscanf(f,"%lg",&(UNOK[l][i]));
    /* Leemos la matriz theta: */
    for(i=0;i<num_reduc_sep[l];i++) {
      for(j=0;j<num_ejemplos_sep[l];j++) {
        fscanf(f,"%lg",&(K[l][i][j]));
      }
    }
  }    
    
  /* Y ahora comenzamos a leer los ejemplos: */
  num_ejemplos_modelo = 0;
  num_dimensiones = 0;
  fscanf(f,"%d",&clase);
  while(1) {
    tabla_modelo[num_ejemplos_modelo].clase = clase;
    for(i=0;i<MAXDIM;i++)
      tabla_modelo[num_ejemplos_modelo].vector[i] = 0.0;   
    while(fscanf(f,"%d:%lg ",&atributo,&valor)==2){
      tabla_modelo[num_ejemplos_modelo].vector[atributo-1] = valor;
      if(atributo>num_dimensiones)
        num_dimensiones=atributo;
    }
    num_ejemplos_modelo++;
    if(feof(f)){
      break;
    }
    clase=atributo;
  }
  fclose(f);
}


double kernel(double *v1,double *v2)
{
  int i;
  double acum=0;

  /* RBF: */  
  for(i=0;i<num_dimensiones;i++) {
    acum += (v1[i]-v2[i])*(v1[i]-v2[i]);
  }
  return exp(-gamm*acum);
}

int KMahClassify(double *v)
{
  double acum[2],temp;
  int i,j,l,jj[2];
  double txmmx[2][MAXEJEMPLOSSEP];
  
  jj[0]=-1;
  jj[1]=-1;
  for(j=0;j<num_ejemplos_modelo;j++) {
    l = (tabla_modelo[j].clase==1?1:0);
    jj[l]++;
    txmmx[l][jj[l]] = kernel(v,tabla_modelo[j].vector) - UNOK[l][jj[l]];
  }

  for(l=0;l<2;l++) {  
    acum[l] = 0;
    for(i=0;i<num_reduc_sep[l];i++) {
      temp = 0;
      for(j=0;j<num_ejemplos_sep[l];j++) {
        temp += K[l][i][j]*txmmx[l][j];
      }  
      acum[l] += temp*temp;
    }
  }
  
  acum[0] = sqrt(acum[0]);
  acum[1] = sqrt(acum[1]);
      
  /* Separador lineal: */
  if(W[0]*acum[0]+W[1]*acum[1]+W[2]>0) {
    return 1;
  } else {
    return -1;
  }
}

