#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <sys/times.h>
#include <time.h>

#include "Kernel_media_learn.h"

TEjemplo tabla_ejemplos[MAXEJEMPLOS];
int num_ejemplos,num_ejemplos_sep[2],num_dimensiones;
double gamm;
double dist[MAXEJEMPLOS][3];
double unokauno[2]; /* Uno para cada clase. */
double XTX[3][3],XTXInv[3][3],XTB[3],W[3]; /* Para el separador lineal */

main(int argc, char *argv[])
{
  int i,j,k,indice,indice1,indice2;
  struct tms buffer_time;
  clock_t tiempo;
  double valdet;
    
  /* Comprobamos correccion de la linea de comandos. */
  if(argc!=4) {
    printf("USO: Kernel_media_learn ancho_rbf(double) fich_entrada fich_salida\n");
    exit(0);
  }

  /* Leemos gamma de la linea de comandos: */
  gamm = (double)atof(argv[1]);

  /*printf("EPSILON=%lf\n",EPSILON);*/
 
  /* Leemos el fichero en memoria: */
  leer_ejemplos(argv[2]);
  printf("%d ejemplos, %d dimensiones\n",num_ejemplos,num_dimensiones);
  /*for(i=0;i<num_ejemplos;i++){
    printf("\nClase:%d - ",tabla_ejemplos[i].clase);
    for(j=0;j<num_dimensiones;j++){
      printf("%lf ",tabla_ejemplos[i].vector[j]);
    }
  }*/

  times(&buffer_time);
  tiempo = buffer_time.tms_utime;
  
  /* Calculamos 1.K.1 (la media de la matriz de productos internos): */
  unokauno[0] = 0;
  unokauno[1] = 0;
  num_ejemplos_sep[0] = 0;
  num_ejemplos_sep[1] = 0;
  for(i=0;i<num_ejemplos;i++) {
    indice1 = (tabla_ejemplos[i].clase==1?1:0); 
    num_ejemplos_sep[indice1]++;
    for(j=0;j<num_ejemplos;j++) {
      indice2 = (tabla_ejemplos[j].clase==1?1:0); 
      if(indice1==indice2) { 
        unokauno[indice1] += kernel(tabla_ejemplos[i].vector,tabla_ejemplos[j].vector);
      }
    }  
  }
  unokauno[0] /= (num_ejemplos_sep[0]*num_ejemplos_sep[0]);
  unokauno[1] /= (num_ejemplos_sep[1]*num_ejemplos_sep[1]);
    
  printf("Fin del calculo de 1.K.1\n");
  
  /* Ahora calculamos las distancias a las medias de todos los ejemplos: */
  for(j=0;j<num_ejemplos;j++) {
    dist[j][0] = 0;
    dist[j][1] = 0;
    for(i=0;i<num_ejemplos;i++) {
      indice = (tabla_ejemplos[i].clase==1?1:0);
      dist[j][indice] += kernel(tabla_ejemplos[j].vector,tabla_ejemplos[i].vector);
    }
    dist[j][0] = kernel(tabla_ejemplos[j].vector,tabla_ejemplos[j].vector)+unokauno[0]-dist[j][0]*2/num_ejemplos_sep[0];
    dist[j][1] = kernel(tabla_ejemplos[j].vector,tabla_ejemplos[j].vector)+unokauno[1]-dist[j][1]*2/num_ejemplos_sep[1];
    dist[j][0] = sqrt(dist[j][0]);
    dist[j][1] = sqrt(dist[j][1]);
    dist[j][2] = 1; /* Componente homognea. */
  }
  
  
  /* Ahora sacamos la frontera lineal por mnimos cuadrados: */
  /* Primero la matriz XT.X: */
  for(i=0;i<3;i++) {
    for(j=i;j<3;j++) {
      XTX[i][j] = 0;
      for(k=0;k<num_ejemplos;k++)
        XTX[i][j] += dist[k][i]*dist[k][j];
      if(i!=j) {
        XTX[j][i] = XTX[i][j];
      }
    }
  }
  /* Inversa de una matrix 3x3: */
  valdet = -XTX[0][2]*XTX[0][2]*XTX[1][1] + 2*XTX[0][1]*XTX[0][2]*XTX[1][2]
           -XTX[0][0]*XTX[1][2]*XTX[1][2] - XTX[0][1]*XTX[0][1]*XTX[2][2] +
            XTX[0][0]*XTX[1][1]*XTX[2][2];
  XTXInv[0][0] = (-XTX[1][2]*XTX[1][2]+XTX[1][1]*XTX[2][2])/valdet;
  XTXInv[0][1] = (XTX[0][2]*XTX[1][2]-XTX[0][1]*XTX[2][2])/valdet;
  XTXInv[0][2] = (-XTX[0][2]*XTX[1][1]+XTX[0][1]*XTX[1][2])/valdet;
  XTXInv[1][0] = XTXInv[0][1];
  XTXInv[1][1] = (-XTX[0][2]*XTX[0][2]+XTX[0][0]*XTX[2][2])/valdet;
  XTXInv[1][2] = (XTX[0][1]*XTX[0][2]-XTX[0][0]*XTX[1][2])/valdet;
  XTXInv[2][0] = XTXInv[0][2];  
  XTXInv[2][1] = XTXInv[1][2];
  XTXInv[2][2] = (-XTX[0][1]*XTX[0][1]+XTX[0][0]*XTX[1][1])/valdet;
  /* Clculo de XT.b :*/    
  for(i=0;i<3;i++) {
    XTB[i] = 0;
    for(j=0;j<num_ejemplos;j++) {
      if(tabla_ejemplos[j].clase==1) {
        XTB[i] += dist[j][i];
      } else {
        XTB[i] -= dist[j][i];
      }
    }
  }
  /* Clculo final de W (W[0],w[1],w[3]=b)*/
  for(i=0;i<3;i++) {
    W[i] = 0;
    for(j=0;j<3;j++) {
      W[i] += XTXInv[i][j]*XTB[j];
    }
  }
  
  times(&buffer_time);

  printf("Tiempo (minutos de CPU)=%2.2f\n",(double)(buffer_time.tms_utime-tiempo)/(CLK_TCK*60));
  
  /* Escribimos la maquina en el fichero: */
  escribir_maquina(argv[3]);
}


void leer_ejemplos(char *fichero)
{
  int i,clase,atributo;
  double valor;
  FILE *f;

  f=fopen(fichero,"r");
  if(f==NULL) {
    perror("Error en apertura de fichero de ejemplos.");
    exit(-1);
  }
  fscanf(f,"%d",&clase);
  num_ejemplos=0;
  num_dimensiones=0;

  while(1) {
    tabla_ejemplos[num_ejemplos].clase = clase;
    for(i=0;i<MAXDIM;i++)
      tabla_ejemplos[num_ejemplos].vector[i] = 0.0;   
    while(fscanf(f,"%d:%lf ",&atributo,&valor)==2){
      tabla_ejemplos[num_ejemplos].vector[atributo-1] = valor;
      if(atributo>num_dimensiones)
        num_dimensiones=atributo;
    }
    num_ejemplos++;
    if(feof(f)){
      break;
    }
    clase=atributo;
  }
  fclose(f);
}

void escribir_maquina(char *fichero)
{
  int i,j;
  FILE *f;

  f=fopen(fichero,"w");
  if(f==NULL) {
    perror("Error en escritura de fichero de maquina.");
    exit(-1);
  }

  /* Escribimos el valor de gamma: */
  fprintf(f,"%lf\n",gamm);
  /* Escribimos el numero de ejemplos y los valores de unokauno: */
  fprintf(f,"%d\n",num_ejemplos);
  fprintf(f,"%lf\n",unokauno[0]);
  fprintf(f,"%lf\n",unokauno[1]);
  /* Escribimos el vector w: */
  for(i=0;i<3;i++) {
    fprintf(f,"%lf ",W[i]);
  }
  fprintf(f,"\n");

  /* Escribimos los ejemplos del modelo: */
  for(i=0;i<num_ejemplos;i++) {
    fprintf(f,"%d ",tabla_ejemplos[i].clase);
    for(j=0;j<num_dimensiones;j++) 
      if(tabla_ejemplos[i].vector[j]!=0.0)
        fprintf(f,"%d:%lf ",j+1,tabla_ejemplos[i].vector[j]);
    fprintf(f,"\n");
  }
  fclose(f);  
}


double kernel(double *v1,double *v2)
{
  int i;
  double acum=0;
  
  /* RBF: */
  for(i=0;i<num_dimensiones;i++) {
    acum += (v1[i]-v2[i])*(v1[i]-v2[i]);
  }
  return exp(-gamm*acum);
  
  /* LINEAL: */
/*  for(i=0;i<num_dimensiones;i++) {
    acum += v1[i]*v2[i];
  }
  return acum;*/
}

